/*============================================================================
Copyright 2006 Thad Harroun, Kevin Yager

This file is part of "yanera 2.0".

"yanera" is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 3 of the License, or (at your option) any later version.

"yanera" is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with "yanera"; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA 
==============================================================================*/ 
/*============================================================================*/
/** \file yanera_xml_input.h
    \brief Parses the XML input file.

  This file handles the loading into memory the XML input file. See the DTD and introduction for complete information on the XML tag semantics. Parsing the XML file requires libxml2 library.
*/
/*============================================================================*/
/** \brief Parses the XML root node.

\param[in] filename The XML input file name
\param yanera A newly allocated \c yanera_container that will hold the data in memory. 

This function handles the root node. It also performs validation agains the DTD.
Children of the root node are:
- \<data_section/\>
- \<model_section/\>
.
 */
void yanera_read_xml(yanera_container *yanera);
/*============================================================================*/
/** \brief Parses the \<model_section/\> node.


\param[in] xml_doc The XML document structure.
\param[in] xml_parent_node In this case the \<model_section/\> of the XML DOM.
\param yanera The \c yanera_container that holds the data in memory.

Children of the \<model_section/\> node are:
- \<model_type/\>
- \<model/\>
- \<component/\>
.
 */
void parseModelSection(xmlDocPtr xml_doc, xmlNodePtr xml_parent_node, \
                       yanera_container *yanera);
/*============================================================================*/
/** \brief Parses the \<model/\> node.


\param[in] xml_doc The XML document structure.
\param[in] xml_node_model The \<model_section/\> node of the XML DOM.
\param model A newly allocated \c yanera_model that will hold the data.
\param yanera The \c yanera_container that holds the data in memory.

The \<model/\> node will form the \c model in memory, so the purpose of the \c yanera variable is access to the \c yanera_parameters. The \c model has already been linked into the list of models in \c yanera.

Children of the \<model/\> node are:
- \<polarized/\>
- \<name/\>
- \<bulk/\>
- \<layer/>
- \<repeat/>
- \<substrate/>
- \<background/>
.
 */
void parseModel(xmlDocPtr xml_doc, xmlNodePtr xml_node_model, \
                yanera_model *model, yanera_container *yanera);
/*============================================================================*/
/** \brief Parses the \<layer/\> node.

\param[in] xml_doc The XML document structure.
\param[in] xml_node_layer This can either be a \<layer/\> node or a \<repeat/\> node.
\param layer A newly allocated \c yanera_layer that will hold the data.
\param yanera The \c yanera_container that holds the data in memory.

The \<layer/\> node will form the \c layer in memory, so the purpose of the \c yanera variable is access to the \c yanera_parameters. The \c layer has already been linked into the list of layers in the appropriate model.

Possible children of the \<layer/\> node are:
- \<type/\>
- \<name/\>
- \<func/\>
- \<parm/\>
- \<thik/\>
- \<sigz/\>
- \<cntr/\>
- \<rsld/\>
- \<isld/\>
- \<rmag/\>
- \<include/\>
.
 */
void parseLayer(xmlDocPtr xml_doc, xmlNodePtr xml_node_layer, \
                yanera_layer *layer, yanera_container *yanera);
/*============================================================================*/
/** \brief Parses a \<repeat/\> node

\param[in] xml_doc The XML document structure.
\param[in] xml_node_parent In this case a \<model/\> node of the XML DOM.
\param[in] repeat_number A property of the \<repeat/\> node indicating the number of types the groupd of layers should be repeated.
\param model A newly allocated \c yanera_model that will hold the data.
\param yanera The \c yanera_container that holds the data in memory.

The function uses a temporary model to hold a copy of the repeating group of layers. It then copies that group into the model the correct number of times.

Children of the \<repeat/\> node are:
- \<layer/>
.
 */
void parseRepeat(xmlDocPtr xml_doc, xmlNodePtr xml_node_parent, \
                 unsigned short repeat_number, \
                 yanera_model *model, yanera_container *yanera);
/*============================================================================*/
/** \brief Parses the \<data_section/\>

\param[in] xml_doc The XML document structure.
\param[in] xml_node_parent In this case the node of the XML DOM.
\param yanera The \c yanera_container that holds the data in memory.

This function does very little at this point. Later versions will load the data into memory.

Children of the \<data_section/\> node are:
- \<data/\>
.
 */
void parseDataSection(xmlDocPtr xml_doc, xmlNodePtr xml_node_parent, \
                          yanera_container *yanera);
/*============================================================================*/
/** \brief Parses a \<data/\> node.


\param[in] xml_doc The XML document structure.
\param[in] xml_node_parent In this case a \<data_section/\> node of the XML DOM.
\param data A newly allcoated \c yanera_data that holds the data in memory.

This function does very little at this point. Later versions will load the data into memory.

Children of the \<data/\> node are:
- \<filename/\>
- \<resolution/\>
.
 */
void parseData(xmlDocPtr xml_doc, xmlNodePtr xml_node_parent, \
                   yanera_data *data);
/*============================================================================*/
/** \brief Parses a \<resolution/\> node.

\param [in] xml_doc The XML document structure.
\param [in] xml_node_data A \<data/\> node of the XML DOM.
\param data A newly allcoated \c yanera_data that holds the data in memory.

This function does very little at this point. It will determine the type of resolution function to apply to the reflectivity calculation.

Children of the \<data/\> node are:
- \<type/\>
.
 */
void parseResolution(xmlDocPtr xml_doc, xmlNodePtr xml_node_data, \
                     yanera_data *data);
/*============================================================================*/
/** \brief Parses a \<miscellaneous_section/\> node.

\param [in] xml_doc The XML document structure.
\param [in] xml_parent_node In this case the \<miscellaneous_section/\> of the XML DOM.
\param yanera The \c yanera_container that holds the data in memory.

This function reads miscellaneous information from the \<miscellaneous_section/\>

Children of the \<miscellaneous_section/\> node are:
- \<write_results/\>
- \<write_profile/\>
- \<write_reflectivity/\>
- \<write_slabs/\>
- \<do_fit/\>
- \<q_min/\>
- \<q_max/\>
- \<q_num/\>
- \<quadrature_error/\>
.
 */
void parseMiscellaneousSection(xmlDocPtr xml_doc, xmlNodePtr xml_parent_node, \
                               yanera_container *yanera);
/*============================================================================*/
/** \brief Parses \<min/\max/\> nodes.

\param [in] xml_doc The XML document structure.
\param [in] xml_node_parent In this case the parameter node currently being read.
\param index The array index of the parameter currently being read.
\param yanera The \c yanera_container that holds the data in memory.

This function reads minimum and maximum constraints of a parameter
 */
void parseMinMax(xmlDocPtr xml_doc, xmlNodePtr xml_node_parent, \
            unsigned short index, yanera_container *yanera);
